import kazoo

from prompt_toolkit.completion import Completer, Completion

from .lexer import KEYWORDS


class ZkCompleter(Completer):

    def __init__(self, zkcli, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.zkcli = zkcli
        self.command = None
        self.prev_typed_word = None
        self.cache = {}

    def get_completions(self, document, complete_event):
        """The completions are either commands or paths"""
        word_before_cursor = document.get_word_before_cursor(WORD=True)

        # If command has been typed, store it
        if (
                not self.command and
                self.prev_typed_word in KEYWORDS and
                word_before_cursor == ''
        ):
            self.command = word_before_cursor

        # Find all possible completable commands
        self.prev_typed_word = word_before_cursor
        if self.command is None:
            completions = [cmd for cmd in KEYWORDS if cmd.startswith(word_before_cursor)]
            for completion in completions:
                yield Completion(completion, -len(word_before_cursor))
        else:
            # Autocomplete on the path of available znodes
            path = word_before_cursor
            if path.startswith('/'):
                current_chroot = '/'.join(path.split('/')[:-1]).rstrip('/') or '/'
                current_node = path.replace(current_chroot, '').lstrip('/')

                if current_chroot not in self.cache:
                    try:
                        self.cache[current_chroot] = self.zkcli.get_children(current_chroot)
                    except kazoo.exceptions.NoNodeError:
                        # We may be typing a nonexistent path
                        # (for example with the 'create' command)
                        self.cache[current_chroot] = []

                completions = [
                    '/%s' % (node)
                    for node in self.cache[current_chroot]
                    if node.startswith(current_node)
                ]

                for completion in completions:
                    yield Completion(completion, -(len(current_node) + 1))



from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError, ConnectionLoss


class ExtendedKazooClient(KazooClient):

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def command(self, cmd='ruok'):
        """Sends a commmand to the ZK node.

        Overrides methode defined at
        https://github.com/python-zk/kazoo/blob/release/2.4/kazoo/client.py#L637
        as it could leave some data unread from the socket.

        """
        if not self._live.is_set():
            raise ConnectionLoss("No connection to server")
        out = []
        peer = self._connection._socket.getpeername()
        if len(peer) > 2:
            peer = peer[:2]
        sock = self.handler.create_connection(
            peer, timeout=self._session_timeout / 1000.0)
        sock.sendall(cmd)
        while True:
            data = sock.recv(8192)
            if not data:
                break
            out.append(data)

        sock.close()
        result = b''.join(out)
        return result.decode('utf-8', 'replace')

    def stat(self, node_path):
        try:
            _, stat = self.get(node_path)
        except NoNodeError:
            return None
        else:
            return stat



import re
import datetime
import functools
import tempfile
import subprocess
import os

import colored
from kazoo.exceptions import NoNodeError, NotEmptyError

from .lexer import COMMAND, PATH, QUOTED_STR, KEYWORDS, FOUR_LETTER_WORD
from .formatting import colorize, columnize, PARENT_ZNODE_STYLE
from .validation import validate_command_input, ask_for_confirmation

# A CLI user-input token can either be a command, a path or a string
TOKEN = r'(%s)' % '|'.join([COMMAND, PATH, QUOTED_STR, FOUR_LETTER_WORD])
NODES_PER_LINE = 3


class UnauthorizedWrite(Exception):
    """Exception raised when a write operation is triggered in RO mode."""


def command_usage(command_name):
    """Print the usage of the argument command."""
    doc = getattr(ZkCommandRunner, command_name).__doc__
    lines = doc.split('\n')[1:]  # remove first line
    lines = [line.strip() for line in lines]
    return '\n'.join(lines).strip()


def command_help(command_name, short=False):
    """Display the help (short or not) of the argument command."""
    command = getattr(ZkCommandRunner, command_name, None)
    if not command:
        return ''
    doc = command.__doc__
    if not doc:
        return ''
    lines = doc.split('\n')
    lines = [line.strip() for line in lines]
    if short:
        lines = [lines[0]]
    return '\n'.join(lines).strip()


def commands_help():
    """Display the short help of all available commands."""
    _help = ['Commands:']
    for command_name in KEYWORDS:
        command_short_help = command_help(command_name, short=True)
        _help.append('- %s: %s' % (command_name, command_short_help))
    return '\n'.join(_help)


def write_op(f):
    """Raise an UnauthorizedWrite exception if the client is in read-only mode."""
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        runner, *_ = args
        if runner.zkcli.read_only:
            raise UnauthorizedWrite('Un-authorized write operation in read-only mode')
        return f(*args, **kwargs)
    return wrapper


class ZkCommandRunner:
    """Object in charge of running the zookeeper commands."""

    def __init__(self, zkcli):
        self.zkcli = zkcli

    def _tokenize(self, command_str):
        tokens = re.findall(TOKEN, command_str)
        return [tok[0].strip() for tok in tokens]

    def exit(self):
        """Close the shell"""
        raise KeyboardInterrupt

    def quit(self):
        """Close the shell"""
        raise KeyboardInterrupt

    def help(self, command_name=None):
        """Print the help of a command

        Usage: help [command]
        Examples:
        - help     # shows the list of commands
        - help ls  # shows a command help

        """
        if command_name:
            return command_help(command_name)
        else:
            return commands_help()

    def ls(self, path):
        """Display the children of a ZNode

        Usage: ls <path>
        Example: ls /test

        """
        try:
            nodes = self.zkcli.get_children(path)
        except NoNodeError as exc:
            raise NoNodeError('%s does not exist' % (path))

        nodes = sorted(nodes)
        fmt_nodes = []
        for node in nodes:
            if self.zkcli.get_children('/'.join((path, node))):
                node = colored.stylize(node, PARENT_ZNODE_STYLE)
            fmt_nodes.append(node)

        nodes = columnize(fmt_nodes, NODES_PER_LINE)
        return nodes

    def _get(self, path):
        try:
            node_data, _ = self.zkcli.get(path)
        except NoNodeError:
            raise NoNodeError('%s does not exist' % (path))
        else:
            if node_data is not None:
                node_data = node_data.decode('utf-8')
                return node_data

    @colorize
    def get(self, path):
        """Display the content of a ZNode

        Usage: get <path>
        Example: get /test

        """
        return self._get(path)

    @write_op
    def create(self, path):
        """Recursively create a path if it doesn't exist

        Usage: create <path>
        Example: create /a/b/c/d

        """
        return self.zkcli.ensure_path(path)

    @write_op
    def set(self, path, data=None):
        """Set or update the content of a ZNode

        Usage: set <path> [data]
        Example: set /test '{"key": "value"}'
                 set /test

        """
        if data:
            data = data.strip("'").strip('"')
        else:
            tmpfd, tmpf_path = tempfile.mkstemp()
            os.close(tmpfd)
            subprocess.call([os.environ.get('EDITOR', 'vim'), tmpf_path])
            data = open(tmpf_path).read()
            os.remove(tmpf_path)

        if not self.zkcli.exists(path):
            return self.zkcli.create(path, data.encode('utf-8'))
        else:
            self.zkcli.set(path, data.encode('utf-8'))

    @write_op
    def edit(self, path):
        """Edit the content of a ZNode

        Usage: edit <path>
        Example: edit /test

        """
        data = self._get(path)

        with tempfile.NamedTemporaryFile(mode='w', encoding='utf-8', delete=False) as f:
            f.write(data)
        subprocess.call([os.environ.get('EDITOR', 'vim'), f.name])
        data = open(f.name).read()
        os.remove(f.name)
        self.zkcli.set(path, data.encode('utf-8'))

    @write_op
    def delete(self, path):
        """Delete a leaf ZNode

        Usage: delete <path>
        Example: delete /test/node

        """
        if ask_for_confirmation("You're about to delete %s. Proceed?" % (path)):
            try:
                self.zkcli.delete(path)
            except NotEmptyError:
                msg = (
                    'Cannot delete %s as it still contains nodes. '
                    'Use the `rmr` command if you really want to proceed.') % (path)
                raise NotEmptyError(msg)
            except NoNodeError:
                raise NoNodeError('%s does not exist' % (path))

    @write_op
    def rmr(self, path):
        """Recursively delete all children ZNodes, along with argument node.

        Usage: rmr <path>
        Example: rmr /test

        """
        if ask_for_confirmation("You're about to recursively delete %s. Proceed?" % (path)):
            self.zkcli.delete(path, recursive=True)

    def stat(self, path):
        """Display a ZNode's metadata

        Usage: stat <path>
        Example: stat /test

        """
        def dtfmt(ts):
            return datetime.datetime.fromtimestamp(ts / 1000).strftime(
                "%a %b %d %H:%M:%S UTC %Y")
        stat = self.zkcli.stat(path)
        lines = [
            'cZxid = {0:x}'.format(stat.czxid),
            'ctime = {}'.format(dtfmt(stat.ctime)),
            'mZxid = {0:x}'.format(stat.mzxid),
            'mtime = {}'.format(dtfmt(stat.mtime)),
            'pZxid = {0:x}'.format(stat.pzxid),
            'cversion = {}'.format(stat.cversion),
            'dataVersion = {}'.format(stat.version),
            'aclVersion = {}'.format(stat.aversion),
            'ephemeralOwner = {0:x}'.format(stat.ephemeralOwner),
            'dataLength = {}'.format(stat.dataLength),
            'numChildren = {}'.format(stat.numChildren),
        ]
        return '\n'.join(lines)

    def toggle_write(self):
        """Activate/deactivate read-only mode

        Usage: toggle_write

        """
        self.zkcli.read_only = not self.zkcli.read_only

    def raw(self, _4lcmd):
        """Send the 4-letter-word command to the zookeeper server

        Usage: raw <4-letter-word>
        Example: raw srvr

        """
        _4lcmd = _4lcmd.encode('utf-8')
        return self.zkcli.command(_4lcmd)

    @validate_command_input
    def run(self, command_str):
        if command_str:
            tokens = self._tokenize(command_str)
            command, args = tokens[0].strip(), tokens[1:]
            out = getattr(self, command)(*args)
            return out



from pygments.lexer import RegexLexer, words
from pygments.token import Keyword, Text, String

KEYWORDS = [
    # 'addauth',
    # 'close',
    # 'connect',
    'create',
    'delete',
    'edit',
    'exit',
    # 'delquota',
    'get',
    # 'getAcl',
    'help',
    # 'history',
    # 'listquota',
    'ls',
    # 'ls2',
    # 'printwatches',
    'quit',
    # 'redo',
    'raw',
    'rmr',
    'set',
    # 'setAcl',
    # 'setquota',
    'stat',
    # 'sync',
    'toggle_write'
]

# A zookeeper CLI command
COMMAND = r'(%s)' % ('|'.join(KEYWORDS))

# A znode path
PATH = r'/[^\s]*'

# A string-value
QUOTED_STR = r"('[^']*'|\"[^\"]*\")"

# A single 4 letter word
FOUR_LETTER_WORD = r'\w{4}'


class ZkCliLexer(RegexLexer):

    tokens = {
        'root': [
            (PATH, Text),
            (FOUR_LETTER_WORD, Text),
            (QUOTED_STR, String),
            (words(KEYWORDS, suffix=r'\b'), Keyword),
        ],
    }



import json
import functools
import shutil

import colored
from pygments import highlight, lexers, formatters
from pygments import styles


STYLE_NAMES = list(styles.get_all_styles()) + ['none']
PARENT_ZNODE_STYLE = angry = colored.fg("blue") + colored.attr("bold")


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def columnize(items, nb_columns):
    """Format the argument items in columns, using the current terminal width"""
    items_lines = chunks(items, nb_columns)
    term_width, _ = shutil.get_terminal_size()
    col_width = int(term_width / nb_columns)
    lines = '\n'.join([
        "".join(item.ljust(col_width) for item in item_line)
        for item_line in items_lines
    ])
    return lines


def colorize(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        """Pretty-print container types and syntax-highlight command results."""
        from .prompt import g
        out = f(*args, **kwargs)
        if out is None:
            return
        if g.style is None:
            return out
        printable = repr(out)
        lexer = lexers.PythonLexer()
        try:
            serialized = json.loads(out)
        except ValueError:
            pass
        else:
            if isinstance(serialized, (tuple, list, dict)):
                printable = json.dumps(serialized, indent=2)
                lexer = lexers.JsonLexer()
        printable = highlight(
            printable,
            lexer,
            formatters.Terminal256Formatter(style=g.style))
        return printable
    return wrapper



import argparse
import threading
import os

from prompt_toolkit.shortcuts import prompt
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
try:
    from prompt_toolkit.lexers import PygmentsLexer
except ImportError:
    def PygmentsLexer(lexer):
        return lexer
try:
    from prompt_toolkit.styles.pygments import style_from_pygments_cls
except ImportError:
    def style_from_pygments_cls(style):
        return style
from pygments import styles
from kazoo.exceptions import NoNodeError, NotEmptyError

from .runner import ZkCommandRunner, command_usage, UnauthorizedWrite
from .lexer import ZkCliLexer
from .zk import ExtendedKazooClient
from .completion import ZkCompleter
from .validation import UnknownCommand, CommandValidationError, ask_for_confirmation
from .formatting import STYLE_NAMES
from .utils import bool_from_str
from . import __version__


history = InMemoryHistory()
auto_suggest = AutoSuggestFromHistory()
g = threading.local()

DEFAULT_ZK_URL = 'localhost:2181'
DEFAULT_COLOR_STYLE = 'monokai'
DEFAULT_INPUT_MODE = 'vi'


class EnvDefault(argparse.Action):
    """Action allowing to define a command-line argument via an environment variable.

    The environment variable must be named IZK_<flag name in uppercase>.

    Example: --write can be overriden via the IZK_WRITE environment variable.

    """
    def __init__(self, required=True, default=None, **kwargs):
        # Read value from environment variable, and apply type, if any
        envvar = 'IZK_' + kwargs['dest'].upper()
        if os.environ.get(envvar) and os.environ[envvar] != default:
            if kwargs.get('type'):
                if kwargs['type'] == bool:
                    default = bool_from_str(os.environ[envvar])
                else:
                    default = kwargs['type'](os.environ[envvar])
            else:
                default = os.environ[envvar]

        # If value was required but the envvar was given, cancel the requirement
        if required and default is not None:
            required = False

        # Update the help string with the override environment variable info
        additional_help = 'Override via the %s environment variable.' % (envvar)
        if 'help' in kwargs:
            kwargs['help'] = '%s. %s' % (kwargs['help'].rstrip('.'), additional_help)
        else:
            kwargs['help'] = additional_help
        super().__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


def parse_args():
    parser = argparse.ArgumentParser(
        description="CLI for zookeeper with syntax-highlighting and auto-completion.",
        epilog="Version: %s" % (__version__))
    parser.add_argument(
        'zk_url',
        nargs='?',
        action=EnvDefault,
        help='URL of the zookeeper node. Default: %s' % (DEFAULT_ZK_URL),
        default=DEFAULT_ZK_URL)
    parser.add_argument(
        '--eval',
        nargs='?',
        type=str,
        help='Evaluate a single zk command and exit')
    parser.add_argument(
        '--write',
        help='Authorize write operations (update/insert/remove)',
        action=EnvDefault,
        type=bool,
        default=False)
    parser.add_argument(
        '--style',
        help="The color style to adopt. Default: %s" % (DEFAULT_COLOR_STYLE),
        action=EnvDefault,
        default=DEFAULT_COLOR_STYLE,
        choices=STYLE_NAMES)
    parser.add_argument(
        '--input-mode',
        help="The input mode to adopt. Default: %s" % (DEFAULT_INPUT_MODE),
        action=EnvDefault,
        default=DEFAULT_INPUT_MODE,
        choices=('vi', 'emacs'))
    parser.add_argument(
        '--version',
        help="Display izk version number and exit",
        action='version',
        version=__version__)
    return parser.parse_args()


def print_headers(zkcli):
    print(zkcli.command(b'srvr'))


def render_prompt(step, read_only):
    mode = 'RO' if read_only else 'RW'
    return '(%s %d) > ' % (mode, step)


def run_cmd(runner, cmd):
    try:
        out = runner.run(cmd)
    except CommandValidationError as exc:
        # The command was invalid. Print command help and usage.
        print(exc, end='\n\n')
        print(command_usage(exc.command))
    except (
        NoNodeError, NotEmptyError, UnknownCommand, UnauthorizedWrite
    ) as exc:
        print(exc)
    else:
        if out is not None:
            print(out)



def main():  # pragma: no cover
    cmd_index = 0
    args = parse_args()
    g.style = None if args.style == 'none' else styles.get_style_by_name(args.style)

    with ExtendedKazooClient(
        hosts=args.zk_url,
        timeout=2,
        read_only=not args.write
    ) as zkcli:
        cmdrunner = ZkCommandRunner(zkcli)
        if args.eval:
            run_cmd(cmdrunner, args.eval)
            return

        print_headers(zkcli)
        while True:
            # We need a new completer for each command
            completer = ZkCompleter(zkcli)
            try:
                cmd = prompt(
                    render_prompt(cmd_index, zkcli.read_only),
                    history=history,
                    auto_suggest=auto_suggest,
                    completer=completer,
                    lexer=PygmentsLexer(ZkCliLexer),
                    style=style_from_pygments_cls(g.style),
                    vi_mode=args.input_mode == 'vi')
                run_cmd(cmdrunner, cmd)
            except (KeyboardInterrupt, EOFError) as exc:
                if ask_for_confirmation('Quit?', confirm_on_exc=True):
                    break
            except Exception:
                raise
            finally:
                cmd_index += 1


if __name__ == '__main__':
    main()



import re

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'izk', '__init__.py')) as init:
    version = re.match(r"__version__ = '(.+)'", init.readline()).group(1)

setup(
    name='izk',
    version=version,
    description='Zookeeper CLI with autocomplete, syntax highlighting and pretty printing',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/brouberol/izk',
    author='Balthazar Rouberol;',
    author_email='br@imap.cc',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='zookeeper cli interactive',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[
        'kazoo',
        'pygments',
        'prompt_toolkit',
        'colored'
    ],
    entry_points={
        'console_scripts': [
            'izk=izk.prompt:main',
        ],
    },
)



import re
import functools

from .lexer import COMMAND, PATH, FOUR_LETTER_WORD, QUOTED_STR


def ask_for_confirmation(message, confirm_on_exc=False):
    try:
        message = '%s [y/n] ' % (message)
        while True:
            answer = input(message)
            if not answer:
                continue
            answer = answer.lower()[0]
            if answer == 'y':
                return True
            elif answer == 'n':
                print('Aborting')
                return False
    except (KeyboardInterrupt, EOFError):
        return confirm_on_exc


class Token(str):
    """A token in a command string"""

    def __init__(self, string):
        self.string = string

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.string)

    def __str__(self):
        return self.string

    def __add__(self, other):
        if isinstance(other, Optional):
            return '%s\s*%s' % (str(self), str(other))
        return '%s\s+%s' % (str(self), str(other))


class Optional(Token):
    """A token that can be ommitted in a command"""

    def __str__(self):
        return r'%s?' % (self.string)


class UnknownCommand(ValueError):
    """Exception raised when an unknown command was passed."""


class CommandValidationError(ValueError):
    """Exception raised when a command received invalid arguments."""

    def __init__(self, command, message):
        self.command = command
        self.message = message

    def __str__(self):
        return self.message


class CommandValidator:
    """Object in charge of validating the user input for a given command."""

    patterns = {
        'create': PATH,
        'delete': PATH,
        'edit': PATH,
        'exit': None,
        'get': PATH,
        'help': Optional(COMMAND),
        'ls': PATH,
        'quit': None,
        'raw': FOUR_LETTER_WORD,
        'rmr': PATH,
        'set': [PATH, Optional(QUOTED_STR)],
        'stat': PATH,
        'toggle_write': None,
    }

    def __init__(self, input_str):
        self.input_str = input_str
        self.command = self._parse_command_name()
        self.pattern = self.make_pattern()

    def _parse_command_name(self):
        command_pattern = r'(%s)\s*' % (COMMAND)
        m = re.match(command_pattern, self.input_str)
        if not m:
            raise UnknownCommand('Command %r not found' % (self.input_str))
        return m.group(1).strip()

    def _make_pattern(self, tokens):
        """Return a regex pattern given the input tokens"""

        # First, transform the input in a list of Tokens, including the command
        if tokens in (None, ''):
            tokens = []
        elif isinstance(tokens, str) and not isinstance(tokens, Token):
            tokens = [Token(tokens)]
        elif isinstance(tokens, list):
            tokens = [tok if isinstance(tok, Token) else Token(tok) for tok in tokens]
        else:
            tokens = [tokens]

        # the first token will be preceeded by the command
        tokens = [Token(self.command)] + tokens

        # Glue the tokens together, and rely on their __add__ method to perform the magic
        pattern = functools.reduce(Token.__add__, tokens)

        return r'^\s*%s\s*$' % (pattern)

    def make_pattern(self):
        tokens = self.patterns[self.command]
        return self._make_pattern(tokens)

    def validate(self):
        return bool(re.match(self.pattern, self.input_str))


def validate_command_input(f):
    """Validate the input command and execute it, if valid"""

    @functools.wraps(f)
    def wrapper(runner, input_str):
        if input_str.strip():
            validator = CommandValidator(input_str.strip())
            if not validator.validate():
                validation_error_msg = 'Command %r is invalid' % (input_str)
                raise CommandValidationError(validator.command, validation_error_msg)
        return f(runner, input_str)
    return wrapper
